/*
 * TalkPeer.java # 2003-9-28
 * mailto: shenchilang@163.com
 */
package cn.nudt.edu.mice.italk.peer;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

/**
 * @author Shen Rui
 * <code>TalkPeer</code> hold a connect to peer
 */
public class TalkPeer implements Runnable, Comparable{
	/**
	 * create a socket with host and port, then
	 * @see TalkPeer(Socket socket)
	 */
	TalkPeer(TalkServer server, String host, int port) throws IOException{
		this(server, new Socket(host, port));
	}

	/**
	 * create a instance with the socket object,
	 * open streams and start the backend thread
	 */
	TalkPeer(TalkServer server, Socket socket) throws IOException{
		_server = server;
		_socket = socket;
		_id = constructId(socket);
		_timestamp = new Date(System.currentTimeMillis());
		_listeners = new HashSet();
		_datas = new HashMap();
		
		// open output stream, since input stream a blocked,
		// we open it at the thread's run() method
		_oos = 
			new ObjectOutputStream(
				_socket.getOutputStream()
			);
		
		// create the thread, and start it right now
		_thread = new Thread(this);
		_thread.start();
	}

	/**
	 * send data to peer, data MUST be serializable 
	 */
	public void send(Object data) throws IOException{
		_oos.writeObject(data);
	}

	public String getName(){
		return _socket.getInetAddress().getHostName();
	}

	public static String constructId(Socket socket){
		InetAddress addr = socket.getInetAddress();
		StringBuffer buf = new StringBuffer();
		buf.append(addr.getHostName())
			.append(" (")
			.append(addr.getHostAddress())
			.append(":")
			.append(socket.getPort())
			.append(":")
			.append(socket.getLocalPort())
			;
		return buf.toString();
	}
	
	public TalkServer getServer(){
		return _server;
	}
	
	/**
	 * get the peer's unique id 
	 */
	public String getId(){
		return _id;
	}

	public void setData(String key, Object data){
		synchronized(_datas){
			_datas.put(key, data);
		}
	}
	
	public Object getData(String key){
		synchronized(_datas){
			return _datas.get(key);
		}
	}

	public boolean isRunning(){
		return _thread != null;
	}
	
	public void addPeerListener(TalkPeerListener listener){
		synchronized(_listeners){
			_listeners.add(listener);
		}
	}
	
	public void removePeerListener(TalkPeerListener listener){
		synchronized(_listeners){
			_listeners.remove(listener);
		}
	}

	/**
	 * close the resources and stop the backend thread
	 */
	public void close(){
		// be sure to close ONLY ONCE
		if (_thread != null){
			_thread = null;
			// this will force the readObject() fail
			try {
				_ois.close();
			} catch (IOException e) {
			}
			try {
				_oos.close();
			} catch (IOException e) {
			}
			try {
				_socket.close();
			} catch (IOException e) {
			}
			_server.disconnect(this);
		}
	}

	/**
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		try{
			// open input stream here, it may block
			_ois = new ObjectInputStream(_socket.getInputStream());
			while(_thread != null){
				Object data = _ois.readObject();
				fireReceived(data);
			}
		}catch(Exception e){
			close();
		}		
	}

	protected void fireReceived(Object data){
		synchronized(_listeners){
			for(Iterator i = _listeners.iterator(); i.hasNext();){
				TalkPeerListener listener = (TalkPeerListener) i.next();
				listener.received(this, data);	
			}
		}
	}
	
	/**
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object that) {
		return compareTo((TalkPeer)that);
	}
	
	private int compareTo(TalkPeer that){
		return 
			this._timestamp
				.compareTo(that._timestamp);
	}

	private TalkServer _server;
	private Socket _socket;
	private String _id;
	private ObjectOutputStream _oos;
	private ObjectInputStream _ois;
	private Thread _thread;
	private Date _timestamp;
	private HashSet _listeners;
	private HashMap _datas;
}
